#ifndef NO_RCSID
const char* qnmungewts_rcsid = 
    "$Header: /u/drspeech/repos/quicknet2/qnmungewts.cc,v 1.4 2014/04/25 03:38:07 davidj Exp $";
#endif

/* Must include the config.h file first */
#include <QN_config.h>
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef QN_HAVE_LIMITS_H
#include <limits.h>
#endif
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)
#define EXIT_FAILURE (1)
#endif
#include <sys/types.h>
#ifdef QN_HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef QN_HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif


#ifndef FILENAME_MAX
#define FILENAME_MAX (MAXPATHLEN)
#endif

#include "QuickNet.h"

static struct {

    const char* in1_file;		// Input weights file.
    const char* in1_format;		// Input weights file format.
    int in1_layers;			// Number of layers to use from in1_file
    const char* in2_file;		// Stacked input weights file.
    const char* in2_format;		// Stacked input weights file format.
    QN_Arg_ListFloat init_random_bias_min; 
    QN_Arg_ListFloat init_random_bias_max;
    QN_Arg_ListFloat init_random_weight_min;
    QN_Arg_ListFloat init_random_weight_max;
    int init_random_seed;	// Seed for random weight initializationx
    QN_Arg_ListInt out_size;	// Output net size
    const char* out_file;		// Output weights file.
    const char* out_format;		// Output weights file format.
    const char* out_output_type;	// Output file output layer type.
    const char* log_file;		// Stream for output data.
    int verbose;		// 1 if verbose.
    int debug;			// Debug level.
} config;

void
set_defaults(void)
{
    static float default_bias_min[2] = { -0.1,-4.1 };
    static float default_bias_max[2] = { 0.1,-3.9 };
    static float default_weight_min[1] = { -0.1 };
    static float default_weight_max[1] = { 0.1 };

    static int default_out_size[1] = { 0 };

    config.in1_file = "";
    config.in1_format = "matlab";
    config.in1_layers = 0;
    config.in2_file = "";
    config.in2_format = "matlab";
    config.init_random_bias_min.count = 2;
    config.init_random_bias_min.vals = &default_bias_min[0];
    config.init_random_bias_max.count = 2;
    config.init_random_bias_max.vals = &default_bias_max[0];

    config.init_random_weight_min.count = 1;
    config.init_random_weight_min.vals = &default_weight_min[0];
    config.init_random_weight_max.count = 1;
    config.init_random_weight_max.vals = &default_weight_max[0];
    config.init_random_seed = 0;
    config.out_size.count = 0;
    config.out_size.vals = &default_out_size[0];
    config.out_file = "";
    config.out_format = "matlab";
    config.out_output_type = "softmax";
    config.log_file = "-";
    config.verbose = 1;
    config.debug = 0;
}

QN_ArgEntry argtab[] =
{
{ NULL, "QuickNet MLP weight file copier version " QN_VERSION, QN_ARG_DESC },
{ "in1_file", "First intput weights file", QN_ARG_STR,
  &(config.in1_file) },
{ "in1_format", "First input weight file format", QN_ARG_STR, 
  &(config.in1_format) }, 
{ "in1_layers", "Number of input layers to use from first input weight file", QN_ARG_INT,
  &(config.in1_layers) },
{ "in2_file", "Stacked input weights file", QN_ARG_STR,
  &(config.in2_file) },
{ "in2_format", "Stacked input weights file format", QN_ARG_STR, 
  &(config.in2_format) }, 
{ "init_random_bias_min", "Minimum random bias (per layer)", QN_ARG_LIST_FLOAT,
  &(config.init_random_bias_min) },
{ "init_random_bias_max", "Maximum random bias (per layer)", QN_ARG_LIST_FLOAT,
  &(config.init_random_bias_max) },
{ "init_random_weight_min", "Minimum random weight (per layer)", QN_ARG_LIST_FLOAT,
  &(config.init_random_weight_min) },
{ "init_random_weight_max", "Maximum random weight (per layer)", QN_ARG_LIST_FLOAT,
  &(config.init_random_weight_max) },
{ "init_random_seed", "Net initialization random number seed",
  QN_ARG_INT, &(config.init_random_seed) },
{ "out_size", "Size of output MLP",
  QN_ARG_LIST_INT, &(config.out_size)},
{ "out_file", "Output weights file", QN_ARG_STR,
  &(config.out_file) },
{ "out_format", "Output weights file format", QN_ARG_STR, 
  &(config.out_format) }, 
{ "out_output_type", "Output wights file output layer type", QN_ARG_STR, 
  &(config.out_output_type) }, 
{ "debug", "Output additional diagnostic information",
  QN_ARG_INT, &(config.debug) },
{ "verbose", "Output diagnostic information",
  QN_ARG_BOOL, &(config.verbose) },
{ NULL, NULL, QN_ARG_NOMOREARGS }
};


const QN_SectionSelector layer2weights[] =
{
    QN_WEIGHTS_NONE,
    QN_LAYER12_WEIGHTS,
    QN_LAYER23_WEIGHTS,
    QN_LAYER34_WEIGHTS,
    QN_LAYER45_WEIGHTS,
    QN_LAYER56_WEIGHTS,
    QN_LAYER67_WEIGHTS,
    QN_LAYER78_WEIGHTS,
    QN_LAYER89_WEIGHTS
};
const QN_SectionSelector layer2biases[] =
{
    QN_WEIGHTS_NONE,
    QN_LAYER2_BIAS,
    QN_LAYER3_BIAS,
    QN_LAYER4_BIAS,
    QN_LAYER5_BIAS,
    QN_LAYER6_BIAS,
    QN_LAYER7_BIAS,
    QN_LAYER8_BIAS,
    QN_LAYER9_BIAS
};


void
qnmungewts(FILE* log_fp)
{
    const int debug = config.debug;
    const int verbose = config.verbose;
    const char* in1_file = config.in1_file;
    const char* in2_file = config.in2_file;
    const char* out_file = config.out_file;
    FILE* in1_file_fp = NULL;
    FILE* in2_file_fp = NULL;
    FILE* out_file_fp = NULL;
    QN_MLPWeightFile* in1_wf = NULL;
    QN_MLPWeightFile* in2_wf = NULL;
    QN_MLPWeightFile* out_wf = NULL;
    QN_MLP* in1_mlp = NULL;	// MLP to put the in1 weights in.
    QN_MLP* in2_mlp = NULL;	// MLP to put the in2 weights in.
    QN_MLP* out_mlp = NULL;	// Output MLP
    size_t in1_layers;
    size_t in1_file_layers = 0;
    size_t in2_file_layers = 0;
    size_t in1_file_layer_size[QN_MLP_MAX_LAYERS];
    size_t in2_file_layer_size[QN_MLP_MAX_LAYERS];
    size_t in_layers = 0;
    size_t in_layer_size[QN_MLP_MAX_LAYERS];
    size_t out_layers;
    size_t out_layer_size[QN_MLP_MAX_LAYERS];
    const QN_OutputLayerType layer_type = QN_OUTPUT_SIGMOID;
    size_t i;
    float min, max;		// Weight limits.

    if (strcmp(in1_file, "")!=0)
    {
	in1_file_fp = QN_open(in1_file, "r");

	const char* in1_format = config.in1_format;
	if (verbose)
	    QN_OUTPUT("Opening input weights file \"%s\"...", in1_file);
	if (strcasecmp(in1_format, "rap3")==0)
	{
	    in1_wf = new QN_MLPWeightFile_RAP3(debug, in1_file_fp,
					       QN_READ,"in1_file", 
					       0, 0, 0);
	}
	else if (strcasecmp(in1_format, "matlab")==0)
	{
	    in1_wf = new QN_MLPWeightFile_Matlab(debug, "in1_file", in1_file_fp,
						 QN_READ, 0, NULL);
	}
	else if (strcasecmp(in1_format, "tnet")==0)
	{
	    in1_wf = new QN_MLPWeightFile_Tnet(debug, "in1_file", in1_file_fp,
						 QN_READ, 0, NULL);
	}
	else
	    QN_ERROR(NULL, "Unknown in1_file file format %s.", in1_format);

	in1_file_layers = in1_wf->num_layers();
	if (config.in1_layers==0)
	    in1_layers = in1_file_layers;
	else
	{
	    if (config.in1_layers<2)
		QN_ERROR(NULL, "in1_layers cannot be less than 2");
	    if ((size_t) config.in1_layers > in1_file_layers)
		QN_ERROR(NULL, "in1_layers (%lu) cannot be greater than number of layers in in1_file (%lu)", (unsigned long)config.in1_layers, (unsigned long) in1_file_layers);
	    in1_layers = (size_t) config.in1_layers;
	}
	for (i = 0; i<in1_file_layers; i++)
	    in1_file_layer_size[i] = in1_wf->size_layer((QN_LayerSelector) i);
	QN_output_mlp_size("in1", in1_file_layers, in1_file_layer_size);
	in_layers = in1_layers;
	for (i = 0; i<in1_layers; i++)
	    in_layer_size[i] = in1_file_layer_size[i];
	QN_output_mlp_size("in1 selected", in1_layers, in_layer_size);
	    

	if (strcmp(in2_file, "")!=0)
	{
	    in2_file_fp = QN_open(in2_file, "r");

	    const char* in2_format = config.in2_format;
	    if (verbose)
		QN_OUTPUT("Opening second input weights file \"%s\"...", in2_file);
	    if (strcasecmp(in2_format, "rap3")==0)
	    {
		in2_wf = new QN_MLPWeightFile_RAP3(debug, in2_file_fp,
						   QN_READ,"in2_file", 
						   0, 0, 0);
	    }
	    else if (strcasecmp(in2_format, "matlab")==0)
	    {
		in2_wf = new QN_MLPWeightFile_Matlab(debug, "in2_file",
						     in2_file_fp,
						     QN_READ, 0, NULL);
	    }
	    else if (strcasecmp(in2_format, "tnet")==0)
	    {
		in2_wf = new QN_MLPWeightFile_Tnet(debug, "in2_file",
						   in2_file_fp,
						   QN_READ, 0, NULL);
	    }
	    else
		QN_ERROR(NULL, "Unknown in2_file file format %s.", in1_format);

	    in2_file_layers = in2_wf->num_layers();
	    for (i = 0; i<in2_file_layers; i++)
		in2_file_layer_size[i] =
		    in2_wf->size_layer((QN_LayerSelector) i);
	    QN_output_mlp_size("in2", in2_file_layers, in2_file_layer_size);
	    
	    if (in2_file_layer_size[0]!=in_layer_size[in1_layers-1])
		QN_ERROR(NULL, "layer %lu of in1_file weights must be same size as first layer of in2_file weights", in1_layers-1);
	    for (i = 0; i<in2_file_layers; i++)
		in_layer_size[in1_layers+i-1] = in2_file_layer_size[i];
	    in_layers += in2_file_layers - 1;

	} // if in2_file specified
	QN_output_mlp_size("in1+in2", in_layers, in_layer_size);
    }
    else // in1_file empty string
    {
	QN_OUTPUT("No input files");
	if (strcmp(in2_file, "")!=0)
	    QN_ERROR(NULL, "in2_file cannot be set without in1_file");
    }

    for (i=0; i<QN_MLP_MAX_LAYERS; i++)
	out_layer_size[i] = 0;
    out_layers = config.out_size.count;
    if (out_layers > 0)
    {
	// Here if user specified output size
	if (out_layers<2)
	    QN_ERROR(NULL, "number of layers for the output file must be at least 2");
	if (out_layers>QN_MLP_MAX_LAYERS)
	    QN_ERROR(NULL, "number of layers for the output file must be at most %lu", QN_MLP_MAX_LAYERS);
	if (out_layers<in_layers)
	    QN_ERROR(NULL, "number of layers for the output file cannot be less than the number of selected input file layers", QN_MLP_MAX_LAYERS);
	for (i = 0; i<out_layers; i++)
	    out_layer_size[i] = config.out_size.vals[i];
	for (i = 0; i<in_layers; i++)
	{
	    if (out_layer_size[i] != in_layer_size[i])
		QN_ERROR(NULL, "layer %lu of combined input nets (%lu units) is a different size from layer %lu of output net (%lu units)", i, in_layer_size[i], i, out_layer_size[i]);
	}
    }
    else
    {
	out_layers = in_layers;
	for (i = 0; i<out_layers; i++)
	    out_layer_size[i] = in_layer_size[i];
    }
    QN_output_mlp_size("out", out_layers, out_layer_size);



    if (strcmp(out_file, "")!=0)
    {
	const char* out_format = config.out_format;
	QN_OutputLayerType type;

	if (strcmp(config.out_output_type, "sigmoid")==0)
	    type = QN_OUTPUT_SIGMOID;
	else if (strcmp(config.out_output_type, "softmax")==0)
	    type = QN_OUTPUT_SOFTMAX;
	else if (strcmp(config.out_output_type, "tanh")==0)
	    type = QN_OUTPUT_TANH;
	else if (strcmp(config.out_output_type, "linear")==0)
	    type = QN_OUTPUT_LINEAR;
	else
	    QN_ERROR(NULL, "Unknown out_output_type layer type %s.",
		     config.out_output_type);

	out_file_fp = QN_open(out_file, "w");
	if (strcasecmp(out_format, "rap3")==0)
	{
	    if (out_layers!=3)
	    {
		QN_ERROR(NULL, "Output weight file has %lu layers but "
			 "RAP3 output file requires 3 layers.",
			 (unsigned long) in1_layers);

	    }
	    else
	    {
		out_wf = new QN_MLPWeightFile_RAP3(debug, out_file_fp,
						   QN_WRITE,"out_file", 
						   out_layer_size[0],
						   out_layer_size[1],
						   out_layer_size[2]);
	    }
	}
	else if (strcasecmp(out_format, "matlab")==0)
	{
	    out_wf = new QN_MLPWeightFile_Matlab(debug, "out_file",
						 out_file_fp,
						 QN_WRITE, out_layers,
						 out_layer_size);
	}
	else if (strcasecmp(out_format, "tnet")==0)
	{
	    out_wf = new QN_MLPWeightFile_Tnet(debug, "out_file",
					       out_file_fp,
					       QN_WRITE, out_layers,
					       out_layer_size,
					       type,
					       0);
	}
	else
	{
	    QN_ERROR(NULL, "Unknown output weight file format %s.",
		     out_format);
	}
    }
    else
    {
	out_file_fp = NULL;
	if (verbose)
	    QN_OUTPUT("No output weight file.");
    }

    if (in1_file_fp!=NULL)
    {
	in1_mlp = new QN_MLP_BunchFlVar(debug, "in1_mlp", in1_file_layers,
					in1_file_layer_size,
					QN_LAYER_SIGMOID, layer_type, 1);
	QN_read_weights(*in1_wf, *in1_mlp, &min, &max, debug, "in1_file");
	if (verbose)
	{
	    QN_OUTPUT("Weights loaded from in1_file '%s', min=%g max=%g.",
		      in1_file, (double) min, (double) max);
	}
    }
    if (in2_file_fp!=NULL)
    {
	in2_mlp = new QN_MLP_BunchFlVar(debug, "in2_mlp", in2_file_layers,
					in2_file_layer_size,
					QN_LAYER_SIGMOID, layer_type, 1);
	QN_read_weights(*in2_wf, *in2_mlp, &min, &max, debug, "in2_file");
	if (verbose)
	{
	    QN_OUTPUT("Weights loaded from in2_file '%s', min=%g max=%g.",
		      in2_file, (double) min, (double) max);
	}
    }


    if (out_wf!=NULL)
    {
	out_mlp = new QN_MLP_BunchFlVar(debug, "out_mlp", out_layers,
					out_layer_size,
					QN_LAYER_SIGMOID, layer_type, 1);

	QN_randomize_weights(debug, config.init_random_seed, *out_mlp,
			     config.init_random_bias_min,
			     config.init_random_bias_max,
			     config.init_random_weight_min,
			     config.init_random_weight_max);
	// Work out a matrix size for copying weights
	size_t max_size = 0;
	for (i=1; i<out_layers; i++)
	{
	    size_t size;
	    
	    size = out_layer_size[i] * out_layer_size[i-1];
	    if (size>max_size)
		max_size = size;
	}
	float* mat = new float [max_size];

	// Read weights and biases from input nets, copying them to output nets
	QN_SectionSelector ibsection, iwsection;
	QN_SectionSelector obsection, owsection;
	QN_MLP* mlp;
	size_t rows, cols;
	for (i=1; i<in_layers; i++)
	{
	    if (i<=in1_layers)
	    {
		ibsection = layer2biases[i];
		iwsection = layer2weights[i];
		mlp = in1_mlp;
	    }
	    else
	    {
		ibsection = layer2biases[i-in1_layers+1];
		iwsection = layer2weights[i-in1_layers+1];
		mlp = in2_mlp;
	    }
	    obsection = layer2biases[i];
	    owsection = layer2weights[i];
	    rows = out_layer_size[i];
	    cols = out_layer_size[i-1];
	    mlp->get_weights(iwsection, 0, 0, rows, cols, mat);
	    out_mlp->set_weights(owsection, 0, 0, rows, cols, mat);
	    mlp->get_weights(ibsection, 0, 0, rows, 1, mat);
	    out_mlp->set_weights(obsection, 0, 0, rows, 1, mat);
	}

	delete[] mat;

	QN_write_weights(*out_wf, *out_mlp, &min, &max, debug, "out_file");
	if (verbose)
	{
	    QN_OUTPUT("Weights written to out_file '%s', min=%g max=%g.",
		      out_file, min, max);
	}
    }
    else
	QN_WARN(NULL, "no output file specified");


    if (out_mlp!=NULL)
	delete out_mlp;
    if (in2_mlp!=NULL)
	delete in2_mlp;
    if (in1_mlp!=NULL)
	delete in1_mlp;
    if (out_wf!=NULL)
	delete out_wf;
    if (in2_wf!=NULL)
	delete in2_wf;
    if (in1_wf!=NULL)
	delete in1_wf;
}

#ifdef NEVER
static void
copy_run(FILE* out_file, FILE* logfile, int verbose)
{

    float min, max;

    QN_read_weights(*inwfile, *mlp, &min, &max, debug, log_fp);
    QN_OUTPUT("Weights loaded from '%s', min=%g max=%g.",
	      init_weight_file, min, max);


    QN_write_weights(*outwfile, *mlp, &min, &max, debug, log_fp);
    QN_OUTPUT("Weights written to '%s', min=%g max=%g.",
	      config.owts_filename, min, max);

    delete mlp;

}

#endif

int
main(int argc, const char* argv[])
{
    char* progname;		// The name of the program - set by QN_args.
    FILE* log_fp;

    set_defaults();
    QN_initargs(&argtab[0], &argc, &argv, &progname);

    log_fp = QN_open(config.log_file, "w");

    if (config.verbose)
	QN_printargs(log_fp, progname, &argtab[0]);
    QN_logger = new QN_Logger_Simple(log_fp, stderr, progname);

    if (config.verbose)
	QN_output_sysinfo("qnmungewts");
    qnmungewts(log_fp);

    delete QN_logger;
    exit(EXIT_SUCCESS);
}
